module dfc.bitmap;

import win32.api;
//import shellapi;
import dfc.core;

import std.c.windows.windows;
import std.c.stdio;

// D foundation Classes
//
// Win32 Bitmap ...
//

/*
 * styles 
 *


 */

class DBitmap
{
public:
	HBITMAP hbm;
	BITMAP bm;

	this()
	{
		hbm = null;
		bm.bmWidth  = -1; 
		bm.bmHeight = -1;
	}

	this( HBITMAP hbmc )
	{
		hbm = hbmc;
		GetObject( hbm, bm.size, &bm );
	}

	this( char[] name )
	{
		this();
		hbm = LoadImage( DFC.globalInstance, name, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE|LR_CREATEDIBSECTION );
		if ( hbm != null )
		{
			GetObject( hbm, bm.size, &bm );
			printf( "image size:(%d, %d)\n", getWidth(), getHeight() );
		}
		else
		{
			printf( "image load failed\n" );
		}
	}

	int getWidth()  { return bm.bmWidth;  }
	int getHeight() { return bm.bmHeight; }

	void renderOnto( HDC dc, int x, int y )
	{
		HDC tmpdc;
		HGDIOBJ oldBitmap;

		tmpdc = CreateCompatibleDC( dc );
		SelectObject( tmpdc, hbm );

		BitBlt( dc, x, y, bm.bmWidth, bm.bmHeight, tmpdc, 0, 0, SRCCOPY );

		DeleteDC( tmpdc );
	}

	void renderSectionOnto( HDC dc, int x, int y, int sx, int sy, int sw, int sh )
	{
		HDC tmpdc;
		HGDIOBJ oldBitmap;

		tmpdc = CreateCompatibleDC( dc );
		oldBitmap = SelectObject( tmpdc, hbm );

		BitBlt( dc, x, y, sw, sh, tmpdc, sx, sy, SRCCOPY );

		SelectObject( tmpdc, oldBitmap );
		DeleteDC( tmpdc );
	}
}

class DCachedBitmap : DBitmap
{
public:
	HDC bmdc;

	this()
	{
		super();
		bmdc = null;
	}

	this( HBITMAP hbmc )
	{
		super( hbmc );
	}

	this( char[] name )
	{
		super( name );
	}

	bit prepareDC( HDC scrndc ) {
		if ( !bmdc ) {
			bmdc = CreateCompatibleDC( scrndc );
			SelectObject( bmdc, hbm );
		}
		return ( bmdc !== null );
	}


	void renderOnto( HDC dc, int x, int y )
	{
		if ( prepareDC( dc ) ) {
			BitBlt( dc, x, y, bm.bmWidth, bm.bmHeight, bmdc, 0, 0, SRCCOPY );
		}
	}

	void renderSectionOnto( HDC dc, int x, int y, int sx, int sy, int sw, int sh )
	{
		if ( prepareDC( dc ) ) {
			BitBlt( dc, x, y, sw, sh, bmdc, sx, sy, SRCCOPY );
		}
	}

	void renderStretchOnto( HDC dc, int x, int y, int w, int h, int sx, int sy, int sw, int sh )
	{
		if ( prepareDC( dc ) ) {
			StretchBlt( dc, x, y, w, h, bmdc, sx, sy, sw, sh, SRCCOPY );
		}
	}
	void renderZoomOnto( HDC dc, int x, int y, int zoom )
	{
		if ( zoom == 0 ) { return; }
		if ( !prepareDC( dc ) ) { return; }
		if ( zoom > 0 ) {
			StretchBlt( dc,   x, y, bm.bmWidth*zoom, bm.bmHeight*zoom,
							bmdc, 0, 0, bm.bmWidth, bm.bmHeight, SRCCOPY );
		} else {
			StretchBlt( dc,   x, y, bm.bmWidth/(-zoom), bm.bmHeight/(-zoom),
							bmdc, 0, 0, bm.bmWidth, bm.bmHeight, SRCCOPY );
		}
	}

	void getZoomParam( int zoom, out int w, out int h ) {
		if ( zoom == 0 ) {
			w = 1;
			h = 1;
		} else if ( zoom > 0 ) {
			w = bm.bmWidth*zoom;
			h = bm.bmHeight*zoom;
		} else {
			w = bm.bmWidth/(-zoom);
			h = bm.bmHeight/(-zoom);
		}
	}

	void renderScaledOnto( HDC dc, int x, int y, int w, int h )
	{
		if ( prepareDC( dc ) ) {
			StretchBlt( dc,   x, y, w, h,
							bmdc, 0, 0, bm.bmWidth, bm.bmHeight, SRCCOPY );
		}
	}
}
