module dfc.button;

import std.c.stdio;

import dfc.core;

// D foundation Classes
//
// Win32 Buttons ...
//

/*
 * WM_COMMAND possible button commands
 *
BN_CLICKED 

// BS_NOTIFY cmd msgs
BN_KILLFOCUS  
BN_SETFOCUS

// 16 bit compatible cmd msg
	--	instead of these you should use BS_OWNERDRAW 
		and process the msg WM_DRAWITEM

BN_DBLCLK / BN_DOUBLECLICKED  
	idButton = (int) LOWORD(wParam);    // identifier of button 
	hwndButton = (HWND) lParam;         // handle to button 
BN_DISABLE  
BN_HILITE / BN_PUSHED
BN_UNHILITE / BN_UNPUSHED
BN_PAINT 

 *
 */

class DButton : DSubClassedWnd
{
public:
	static const char[] NATIVE_CLASS_NAME = "BUTTON";
//	this() { super(Win32.BUTTON); } 
	this() 
	{ 
		super( NATIVE_CLASS_NAME ); 
		addHandler( BN_CLICKED, &onButtonClick );
	} 

	LRESULT onButtonClick()
	{
		printf( "button clicked :)\n" );
		return cast(LRESULT)0;
	}

	void GetPreferedSize( out SIZE size )
	{
		RECT rect;
		HDC dc;
		char[] tmp;
		dc = GetDC( hwnd );
		getWindowText( tmp );

//		HGDIOBJ font = GetStockObject( DEFAULT_GUI_FONT );
		HGDIOBJ font = GetStockObject( SYSTEM_FONT );
		HGDIOBJ oldFont = SelectObject( dc, font );
		GetTextExtentPoint32( dc, tmp, tmp.length, &size );
		SelectObject( dc, oldFont );
		ReleaseDC( hwnd, dc );

		rect.left   = 0;
		rect.top    = 0;
		rect.right  = size.cx+16;
		rect.bottom = size.cy+6;
		adjustRect( rect, false );

		size.cx = rect.right - rect.left;
		size.cy = rect.bottom - rect.top;
	}

	void sendClickMsg()
	{
		sendMsg( BM_CLICK, 0, 0 );
	}

	int getCheck()
	{
		return cast(int)sendMsg( BM_GETCHECK, 0, 0 );
	}

	void setCheck( int checkValue )
	{
		sendMsg( BM_SETCHECK, cast(WPARAM)checkValue, 0 );
	}

	int getState()
	{
		return cast(int)sendMsg( BM_GETSTATE, 0, 0 );
	}
	
	void setState( int state )
	{
		sendMsg( BM_SETSTATE, cast(WPARAM)state, 0 );
	}

	HANDLE getImage( bit asIcon )
	{
		return cast(HANDLE)sendMsg( BM_GETIMAGE, asIcon ? IMAGE_ICON : IMAGE_BITMAP, 0 );
	}

	void setImage( HANDLE image, bit asIcon )
	{
		sendMsg( BM_SETIMAGE, asIcon ? IMAGE_ICON : IMAGE_BITMAP, cast(LPARAM)image );
	}

	void setStyle( uint style, bit redraw )
	{
		sendMsg( BM_SETSTYLE, style, redraw ? 1 : 0 );
	}
}


