module dfc.edit;

import std.c.stdio;
import dfc.core;

// D foundation Classes
//
// Win32 Buttons ...
//

/*
 * WM_COMMAND possible button commands
 *
EN_SETFOCUS         =0x0100,
EN_KILLFOCUS        =0x0200,
EN_CHANGE           =0x0300,
EN_UPDATE           =0x0400,
EN_ERRSPACE         =0x0500,
EN_MAXTEXT          =0x0501,
EN_HSCROLL          =0x0601,
EN_VSCROLL          =0x0602

 *
 */

class DEdit : DSubClassedWnd
{
public:
	static const char[] NATIVE_CLASS_NAME = "EDIT";
//	this() { super(Win32.EDIT); } 
	this() 
	{ 
		super( NATIVE_CLASS_NAME ); 
	} 

//	void setStyle( uint style, bit redraw )
//	{
//		sendMsg( BM_SETSTYLE, style, redraw ? 1 : 0 );
//	}

	LRESULT getSelection( DWORD * startPtr, DWORD * endPtr )
	{
		return sendMsg( EM_GETSEL, cast(WPARAM)startPtr, cast(LPARAM)endPtr );
	}

	LRESULT getSelection( inout DWORD start, inout DWORD end )
	{
		return getSelection( &start, &end );
	}

	void setSelection( DWORD start, DWORD end )
	{
		sendMsg( EM_SETSEL, cast(WPARAM)start, cast(LPARAM)end );
	}

	void getRect( RECT * rectPtr )
	{
		sendMsg( EM_GETRECT, 0, cast(LPARAM)rectPtr );
	}

	void getRect( inout RECT rect )
	{
		getRect( &rect );
	}

	void setRect( RECT * rectPtr, bit redraw )
	{
		sendMsg( redraw ? EM_SETRECT : EM_SETRECTNP, 0, cast(LPARAM)rectPtr );
	}
 
	void setRect( inout RECT rect, bit redraw  )
	{
		setRect( &rect, redraw );
	}

	void setRect( inout RECT rect )
	{
		setRect( &rect, true );
	}
	
	void scroll( int how )
	{
		sendMsg( EM_SCROLL, cast(WPARAM)how, 0 );
	}

	void lineScroll( int cx, int cy )
	{
		sendMsg( EM_LINESCROLL, cast(WPARAM)cx, cast(WPARAM)cy );
	}

	void scrollCaretIntoView()
	{
		sendMsg( EM_SCROLLCARET, 0, 0 );
	}

	bit getModify()
	{
		return (sendMsg( EM_GETMODIFY, 0, 0 )!=0);
	}

	void setModify( bit modified )
	{
		sendMsg( EM_SETMODIFY, cast(WPARAM)modified, 0 );
	}

	int getLineCount()
	{
		return sendMsg( EM_GETLINECOUNT, 0, 0 );
	}

	int getLineIndex( int line )
	{
		return sendMsg( EM_LINEINDEX, cast(WPARAM)line, 0 );
	}

	void setHandle( HLOCAL memHandle )
	{
		sendMsg( EM_SETHANDLE, cast(WPARAM)memHandle, 0 );
	}

	HLOCAL setHandle()
	{
		return cast(HLOCAL)sendMsg( EM_GETHANDLE, 0, 0 );
	}

	int getThumbPos()
	{
		return sendMsg( EM_GETTHUMB, 0, 0 );
	}

	int getLineLength( int charIndex )
	{
		return sendMsg( EM_LINELENGTH, cast(WPARAM)charIndex, 0 );
	}

	version(UNICODE) {
		void replaceSelection( wchar[] newText, bit canUndo )
		{
			wchar * textPtr = cast(wchar*)newText;
			sendMsg( EM_REPLACESEL, cast(WPARAM)canUndo, cast(LPARAM)textPtr );
		}
	} else {
		void replaceSelection( char[] newText, bit canUndo )
		{
			char * textPtr = cast(char*)newText;
			sendMsg( EM_REPLACESEL, cast(WPARAM)canUndo, cast(LPARAM)textPtr );
		}
	}

	version(UNICODE) {
		void getLine( int line, inout wchar[] text, WORD maxLength )
		{ 
			wchar * textPtr = cast(wchar*)text;
			*(cast(WORD*)textPtr) = maxLength;
			sendMsg( EM_REPLACESEL, cast(WPARAM)line, cast(LPARAM)textPtr );
		}
	} else {
		void getLine( int line, inout char[] text, WORD maxLength )
		{ 
			char * textPtr = cast(char*)text;
			*(cast(WORD*)textPtr) = maxLength;
			sendMsg( EM_REPLACESEL, cast(WPARAM)line, cast(LPARAM)textPtr );
		}
	}

	bit canUndo()
	{
		return (sendMsg( EM_CANUNDO, 0, 0 )!=0);
	}

	void undo()
	{
		sendMsg( EM_UNDO, 0, 0 );
	}

	void setLineBreakFormat( bit fmt )
	{
		sendMsg( EM_FMTLINES, cast(WPARAM)fmt, 0 );
	}

	int lineIndexFromCharIndex( int charIndex )
	{
		return sendMsg( EM_LINEFROMCHAR, cast(WPARAM)charIndex, 0 );
	}

// !!! not written this YET 
// EM_SETTABSTOPS

	version(UNICODE) {
		void setPasswordChar( wchar c )
		{
			sendMsg( EM_SETPASSWORDCHAR, cast(WPARAM)c, 0 );
		}
	} else {
		void setPasswordChar( char c )
		{
			sendMsg( EM_SETPASSWORDCHAR, cast(WPARAM)c, 0 );
		}
	}
	
	void clearPasswordMode()
	{
		sendMsg( EM_SETPASSWORDCHAR, 0, 0 );
	}

	void setPasswordMode( bit pwMode )
	{
		sendMsg( EM_SETPASSWORDCHAR, pwMode ? cast(WPARAM)'*' : 0, 0 );
	}

	void emptyUndoBuffer()
	{
		sendMsg( EM_EMPTYUNDOBUFFER, 0, 0 );
	}
	
	int getFirstVisibleLine()
	{
		return sendMsg( EM_GETFIRSTVISIBLELINE, 0, 0 );
	}

	int setReadOnly( bit readonly )
	{
		return sendMsg( EM_SETREADONLY, cast(WPARAM)readonly, 0 );
	}

	EDITWORDBREAKPROC getWordBreakProc()
	{
		return cast(EDITWORDBREAKPROC )sendMsg( EM_GETWORDBREAKPROC, 0, 0 );
	}

	version(UNICODE) {
		wchar setPasswordChar()
		{
			return sendMsg( EM_GETPASSWORDCHAR, 0, 0 );
		}
	} else {
		char setPasswordChar()
		{
			return sendMsg( EM_GETPASSWORDCHAR, 0, 0 );
		}
	}

	bit inPasswordMode()
	{
		return (sendMsg( EM_GETPASSWORDCHAR, 0, 0 )!=0);
	}

// !!! TO sort out later
//EM_SETMARGINS           =0x00D3,
//EM_GETMARGINS           =0x00D4,

	void setTextLimit( int maxLen )
	{
		sendMsg( EM_SETLIMITTEXT, cast(WPARAM)maxLen , 0 );
	}

	int getTextLimit()
	{
		return sendMsg( EM_GETLIMITTEXT, 0, 0 );
	}

	// need to sort out the DWORD<->POINT code

//	LRESULT positionFromCharIndex( int charIndex )
//	{
//		return sendMsg( EM_POSFROMCHAR, cast(WPARAM)charIndex, 0 );
//	}

//	LRESULT charIndexFromPosition( LPARAM position )
//	{
//		return sendMsg( EM_CHARFROMPOS, 0, position );
//	}
}


/*
enum {
// * Edit Control Styles
ES_LEFT             =0x0000,
ES_CENTER           =0x0001,
ES_RIGHT            =0x0002,
ES_MULTILINE        =0x0004,
ES_UPPERCASE        =0x0008,
ES_LOWERCASE        =0x0010,
ES_PASSWORD         =0x0020,
ES_AUTOVSCROLL      =0x0040,
ES_AUTOHSCROLL      =0x0080,
ES_NOHIDESEL        =0x0100,
ES_OEMCONVERT       =0x0400,
ES_READONLY         =0x0800,
ES_WANTRETURN       =0x1000,
ES_NUMBER           =0x2000
}

enum {
// * Edit Control Notification Codes
EN_SETFOCUS         =0x0100,
EN_KILLFOCUS        =0x0200,
EN_CHANGE           =0x0300,
EN_UPDATE           =0x0400,
EN_ERRSPACE         =0x0500,
EN_MAXTEXT          =0x0501,
EN_HSCROLL          =0x0601,
EN_VSCROLL          =0x0602
}

enum {
// Edit control EM_SETMARGIN parameters 
EC_LEFTMARGIN       =0x0001,
EC_RIGHTMARGIN      =0x0002,
EC_USEFONTINFO      =0xffff
}

enum {
// * Edit Control Messages
EM_GETSEL               =0x00B0,
EM_SETSEL               =0x00B1,
EM_GETRECT              =0x00B2,
EM_SETRECT              =0x00B3,
EM_SETRECTNP            =0x00B4,
EM_SCROLL               =0x00B5,
EM_LINESCROLL           =0x00B6,
EM_SCROLLCARET          =0x00B7,
EM_GETMODIFY            =0x00B8,
EM_SETMODIFY            =0x00B9,
EM_GETLINECOUNT         =0x00BA,
EM_LINEINDEX            =0x00BB,
EM_SETHANDLE            =0x00BC,
EM_GETHANDLE            =0x00BD,
EM_GETTHUMB             =0x00BE,
EM_LINELENGTH           =0x00C1,
EM_REPLACESEL           =0x00C2,
EM_GETLINE              =0x00C4,
EM_LIMITTEXT            =0x00C5,
EM_CANUNDO              =0x00C6,
EM_UNDO                 =0x00C7,
EM_FMTLINES             =0x00C8,
EM_LINEFROMCHAR         =0x00C9,
EM_SETTABSTOPS          =0x00CB,
EM_SETPASSWORDCHAR      =0x00CC,
EM_EMPTYUNDOBUFFER      =0x00CD,
EM_GETFIRSTVISIBLELINE  =0x00CE,
EM_SETREADONLY          =0x00CF,
EM_SETWORDBREAKPROC     =0x00D0,
EM_GETWORDBREAKPROC     =0x00D1,
EM_GETPASSWORDCHAR      =0x00D2,
EM_SETMARGINS           =0x00D3,
EM_GETMARGINS           =0x00D4,
EM_SETLIMITTEXT         =EM_LIMITTEXT,   // ;win40 Name change 
EM_GETLIMITTEXT         =0x00D5,
EM_POSFROMCHAR          =0x00D6,
EM_CHARFROMPOS          =0x00D7
}

*/
