module dfc.msg;

import std.c.stdio;


// D foundation Classes
//
// This is a WIN32 GUI library, 
// current plans are that is should be close to the metal
// so there is not Wrapper for a HDC because that would
// just add code that is not required.
//
// This library just wrappers HWND and windows Message proccesing
//
// one day may be a cross platfrom GUI overlayed wiht more classes
// a bit like Java AWT and Swing.
//
import win32.api;

class MessageSlot
{
public:
	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		return false;
	}
}

alias LRESULT delegate(WPARAM,LPARAM) nativeMsgFuncPtr;
class NativeMessageSlot : MessageSlot
{
protected:
	nativeMsgFuncPtr nativeMethod;
public:
	this( nativeMsgFuncPtr toCall )
	{
		nativeMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		rv = nativeMethod( wparam, lparam );
		return true;
	}
}

alias LRESULT delegate(int,int,int) mouseMsgFuncPtr;
class MouseMessageSlot : MessageSlot
{
protected:
	mouseMsgFuncPtr mouseMethod;
public:
	this( mouseMsgFuncPtr toCall )
	{
		mouseMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		rv = mouseMethod( cast(short)LOWORD(lparam), cast(short)HIWORD(lparam), wparam );
		return true;
	}
}

class MouseDownMessageSlot : MouseMessageSlot
{
protected:
	HWND captureWnd;
public:
	this( HWND hwnd, mouseMsgFuncPtr toCall )
	{
		super( toCall );
		captureWnd = hwnd;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		SetCapture( captureWnd );
		rv = mouseMethod( cast(short)LOWORD(lparam), cast(short)HIWORD(lparam), wparam );
		return true;
	}
}

class MouseUpMessageSlot : MouseMessageSlot
{
public:
	this( mouseMsgFuncPtr toCall )
	{
		super( toCall );
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		ReleaseCapture();
		rv = mouseMethod( cast(short)LOWORD(lparam), cast(short)HIWORD(lparam), wparam );
		return true;
	}
}

alias LRESULT delegate(HDC) paintMsgFuncPtr;
class PaintMessageSlot : MessageSlot
{
protected:
	HWND hwnd;
	paintMsgFuncPtr paintMethod;
public:
	this( HWND hwnd0, paintMsgFuncPtr toCall )
	{
		hwnd = hwnd0;
		paintMethod = toCall;
	}
	void setHwnd( HWND hwnd0 ) { hwnd = hwnd0; } 

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		PAINTSTRUCT ps;
		HDC dc;
		dc = BeginPaint( hwnd, &ps );
		rv = paintMethod( dc );
		EndPaint( hwnd, &ps );
		return true;
	}
}

// handler for a CMD message
alias LRESULT delegate() nativeCmdFuncPtr;

alias LRESULT delegate(int scanCode, int keyCode, int count, int flags ) keyMsgFuncPtr;

class KeyMessageSlot : MessageSlot
{
protected:
	keyMsgFuncPtr keyMethod;
public:
	this( keyMsgFuncPtr toCall )
	{
		keyMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		rv = keyMethod( (lparam>>16)&0xFF, wparam, (lparam & 0xFFFF), lparam );
		return true;
	}
}
alias bit delegate(out LRESULT,int,int,int) mouseOptionalMsgFuncPtr;

class MouseOptionalMessageSlot : MessageSlot
{
protected:
	mouseOptionalMsgFuncPtr mouseMethod;
public:
	this( mouseOptionalMsgFuncPtr toCall )
	{
		mouseMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		return mouseMethod( rv, cast(short)LOWORD(lparam), cast(short)HIWORD(lparam), wparam );
	}
}

// RV, width, height, options
alias bit delegate( out LRESULT, int, int, int) sizeOptionalMsgFuncPtr;
class SizeOptionalMessageSlot : MessageSlot
{
protected:
	sizeOptionalMsgFuncPtr sizeMethod;
public:
	this( sizeOptionalMsgFuncPtr toCall )
	{
		sizeMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		return sizeMethod( rv, cast(short)LOWORD(lparam), cast(short)HIWORD(lparam), wparam );
	}
}

alias LRESULT delegate( int, int ) resizeMsgFuncPtr;
class ResizeMessageSlot : MessageSlot
{
protected:
	resizeMsgFuncPtr resizeMethod;
public:
	this( resizeMsgFuncPtr toCall )
	{
		resizeMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		switch( wparam ) {
		case SIZE_MAXIMIZED:
		case SIZE_RESTORED:
			rv = resizeMethod( cast(short)LOWORD(lparam), cast(short)HIWORD(lparam) );
			return true;
		default:
			return false;
		}
	}
}

alias LRESULT delegate( int, int, HWND ) scrollMsgFuncPtr;
class ScrollMessageSlot : MessageSlot
{
protected:
	scrollMsgFuncPtr scrollMethod;
public:
	this( scrollMsgFuncPtr toCall )
	{
		scrollMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		HWND hwnd = cast(HWND)lparam;
//		printf("{pm:in 0x%x\n", wparam );
		rv = scrollMethod( cast(short)LOWORD(wparam), cast(short)HIWORD(wparam), hwnd );
//		printf("}pm:out\n");
		return true;
	}
}

alias LRESULT delegate( HWND ) focusMsgFuncPtr;
class FocusMessageSlot : MessageSlot
{
protected:
	focusMsgFuncPtr focusMethod;
public:
	this( focusMsgFuncPtr toCall )
	{
		focusMethod = toCall;
	}

	bit processMessage( out LRESULT rv, WPARAM wparam, LPARAM lparam )
	{
		HWND hwnd = cast(HWND)wparam;
		rv = focusMethod( hwnd );
		return true;
	}
}

