module dfc.mswin;

// debugging only
import std.c.stdio;

//import dfc.core;

// D foundation Classes
//
// This is a WIN32 GUI library,
// current plans are that is should be close to the metal
// so there is not Wrapper for a HDC because that would
// just add code that is not required.
//
// This library just wrappers HWND and windows Message proccesing
//
// one day may be a cross platfrom GUI overlayed with more classes
// a bit like Java AWT and Swing.
//

import win32.api;
import win32.commctrl;
import dfc.msg;

class DFC
{
public:
	static HINSTANCE globalInstance;
	static const char[] WINDCLASS = "DWindowClass00";
	static const char[] WINDPROP  = "DWindowProp00";

protected:
	static ATOM windPropAtom;
	static Win32Window[HWND] keepalive;
	struct LastItemCache {
		HWND hwnd;
		Win32Window wnd;
	}
	static LastItemCache lastCmd;
	static LastItemCache lastWnd;
	static LastItemCache lastNfy;

	extern(Windows) static LRESULT 
	WindowProc( HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam )
	{
		LRESULT rv;
		Win32Window wnd;
//		printf( "-+- hwnd:%X msg%x(%.*s) [%x %x]\n", cast(int)hwnd, msg, getMsgName_prv( msg ), wparam, lparam );
		// check for NOTIFYS from controls
		if ( msg != WM_NOTIFY )
		{
			// check for COMMANDS from controls
			if ( msg == WM_COMMAND )
			{	
				HWND cmdHwnd = cast(HWND)lparam;
				if ( cmdHwnd == lastCmd.hwnd ) 
				{ 
					wnd = lastCmd.wnd; 
				}
				else
				{
					wnd = windowFromHandle( cmdHwnd );
					lastCmd.hwnd = cmdHwnd;
					lastCmd.wnd  = wnd;
				}
				if ( wnd )
				{
					if ( wnd.handleCommandMsg( rv, HIWORD(wparam) ) )
					{
						return rv;
					}
				}
				// sending the window that windows sent the message to
				// this will most likely just call unhandledMsg
				// but you can add your own command handler if
				// you are putting non DFC attached Windows
				// into your hierachy (not realy advised).
			}

			if ( hwnd == lastWnd.hwnd )
			{
				wnd = lastWnd.wnd;
			}
			else
			{
				wnd = windowFromHandle( hwnd );
				if ( wnd )
				{
					lastWnd.hwnd = hwnd;
					lastWnd.wnd  = wnd;
				}
			}
			if ( wnd )
			{
				// WM_DESTROY comes before WM_NCDESTROY
				// this SHOULD be the last every msg the window will get
				if ( msg == WM_NCDESTROY )
				{
					if(hwnd in keepalive)
					{
						// keepalive[hwnd] = null;
						delete keepalive[hwnd];
					}
				}
				if ( wnd.handleMessage( rv, msg, wparam, lparam ) )
				{
					return rv;
				}
				return wnd.unhandledMsg( hwnd, msg, wparam, lparam);
			}
		}
		else
		{
			LPNMHDR ptrToNotify;
			ptrToNotify = cast(LPNMHDR)lparam;
			if( ptrToNotify )
			{
				if ( ptrToNotify.hwndFrom == lastNfy.hwnd )
				{
					wnd = lastNfy.wnd;
				}
				else
				{
					wnd = windowFromHandle( ptrToNotify.hwndFrom );
					if ( wnd )
					{
						lastNfy.wnd = wnd;
						lastNfy.hwnd = ptrToNotify.hwndFrom;
					}
				}
				if ( wnd )
				{
					if ( wnd.handleNotifyMsg( rv, ptrToNotify ) )
					{
						return rv;
					}
				}
			}
		}

		return DefWindowProcA( hwnd, msg, wparam, lparam);
	}

public:
	static void beginApplication( bit erasebg )
	{
		WNDCLASS wc;
		globalInstance = GetModuleHandleA(null);
		wc.lpszClassName = WINDCLASS;
		wc.style = CS_HREDRAW | CS_VREDRAW;// | CS_PARENTDC; // | CS_OWNDC;

		wc.lpfnWndProc = cast(WNDPROC)(&WindowProc);

		wc.hInstance = globalInstance;
		wc.hIcon = LoadIconA(cast(HINSTANCE) null, cast(char *)IDI_APPLICATION);
		wc.hCursor = LoadCursorA(cast(HINSTANCE) null, cast(char *)IDC_CROSS);
		wc.hbrBackground = erasebg ? cast(HBRUSH) (COLOR_WINDOW + 1) : null;
		wc.lpszMenuName = null;
		wc.cbClsExtra = wc.cbWndExtra = 0;
		assert( RegisterClassA(&wc) );
		windPropAtom = GlobalAddAtomA( WINDPROP );

		// just in case we are using them
		InitCommonControls();
	}

	static void beginApplication()
	{
		beginApplication( false ); // default is not to erase the bg
	}

	static void attach( Win32Window wnd, HWND hwnd )
	{
		keepalive[hwnd] = wnd;
		SetProp( hwnd, cast(char*)windPropAtom, cast(HANDLE)(wnd) );
	}

	static Win32Window windowFromHandle( HWND hwnd )
	{
		if ( hwnd )
		{
			return cast(Win32Window)GetProp( hwnd, cast(char*)windPropAtom );
		}
		return null;
	}

	static WNDPROC subclass( HWND hwnd )
	{
		return cast(WNDPROC)SetWindowLong( hwnd, GWL_WNDPROC, cast(LONG)(&WindowProc) );
	}

	static void msgLoop()
	{
		MSG msg;
		while ( GetMessage( &msg, cast(HWND)null, 0, 0 ) )
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
		GlobalDeleteAtom( windPropAtom );
	}

	/*! check for message and flush, then return 
	*/
	static bit checkForMsgs()
	{
		bit doExit = false;
		MSG msg;
		while ( PeekMessage( &msg, cast(HWND)null, 0, 0, PM_REMOVE ) )
		{
			if ( msg.message == WM_QUIT )
			{
				GlobalDeleteAtom( windPropAtom );
				doExit = true;
			}
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
		return doExit;
	}

	LRESULT sendBroadcastMsg( UINT msg, WPARAM wparam, LPARAM lparam )
	{
		// sends and waits for a responce
		return SendMessage( HWND_BROADCAST, msg, wparam, lparam );
	}

	void postBroadcastMsg( UINT msg, WPARAM wparam, LPARAM lparam )
	{
		PostMessage( HWND_BROADCAST, msg, wparam, lparam );
	}
}


class Win32Window
{
protected:
	char[] nativeClassname;
	HWND hwnd;
	MessageSlot[UINT] msgHandlers;
	// should this be here ?
	nativeCmdFuncPtr[WORD] cmdHandlers;

public:
	this( char[] classname )
	{
		hwnd = null;
		nativeClassname = classname;
	}

	HWND getHwnd() { return hwnd; }
	void CreateWindow(DWORD dwStyle, 
								char[] lpWindowName,
								DWORD dwExStyle,
								int X,
								int Y,
								int nWidth,
								int nHeight,
								HWND hWndParent,
								HMENU hMenu )
	{
		preCreateWindow( dwStyle, dwExStyle );
		hwnd = CreateWindowExA(	dwExStyle,
								cast(char *)nativeClassname,
								cast(char *)lpWindowName,
								dwStyle,
								X, Y,
								nWidth, nHeight,
								hWndParent,
								hMenu,
								DFC.globalInstance,
								null );
		DFC.attach( this, hwnd ); 
		postCreateWindow();
	}

	void preCreateWindow( inout DWORD dwStyle, inout DWORD dwExStyle )
	{
	}

	void postCreateWindow()
	{
	}

	void addHandler( UINT msg, MessageSlot ms )
	{
		msgHandlers[msg] = ms;
	}

	bit handleMessage( out LRESULT rv, UINT msg, WPARAM wparam, LPARAM lparam )
	{
		MessageSlot ms;
		ms = msgHandlers[msg];
		if( ms )
		{
//			printf( " +- hwnd:%X msg%x(%.*s) [%x %x]\n", cast(int)hwnd, msg, getMsgName_prv( msg ), wparam, lparam );
			return ms.processMessage( rv, wparam, lparam );
		}
//		printf( " *\n" );
		return false;
	}

	void addHandler( WORD cmd, nativeCmdFuncPtr cmdEvtFunc )
	{
		cmdHandlers[cmd] = cmdEvtFunc;
	}

	bit handleCommandMsg( out LRESULT rv, WORD cmd )
	{
		nativeCmdFuncPtr cmdSlot = cmdHandlers[cmd];
		if ( cmdSlot )
		{
			rv = cmdSlot();
			return true;
		}
		return false;
	}

	LRESULT unhandledMsg( HWND msghwnd, UINT msg, WPARAM wparam, LPARAM lparam )
	{
		return DefWindowProcA( msghwnd, msg, wparam, lparam);
	}

	bit handleNotifyMsg( out LRESULT rv, LPNMHDR notifyPtr )
	{
		// this will work like the command handlers by checking
		// notifyPtr.code
		// then calling a translation class that converts the
		// LPNMHDR to the required controls notify struct
		return false;
	}

	LRESULT sendMsg( UINT msg, WPARAM wparam, LPARAM lparam )
	{
		int i = cast(int)hwnd;
		// sends and waits for a responce
		LRESULT rv = SendMessage( hwnd, msg, wparam, lparam );
		return rv;
	}

	void postMsg( UINT msg, WPARAM wparam, LPARAM lparam )
	{
		// sends and returns without waiting
		PostMessage( hwnd, msg, wparam, lparam );
	}

	bit sendMsgAsync( UINT msg, WPARAM wparam, LPARAM lparam, SENDASYNCPROC callBack, DWORD data )
	{
		// sends and waits for a responce
		return (SendMessageCallback( hwnd, msg, wparam, lparam, callBack, data )!=0);
	}

	LRESULT sendMsgTimeout( UINT msg, WPARAM wparam, LPARAM lparam, UINT flags, UINT timeout )
	{
		DWORD result = 0;
		SendMessageTimeout( hwnd, msg, wparam, lparam, flags, timeout, &result );
		return cast(LRESULT)result;
	}

	void showWindow( bit state )
	{
		ShowWindow( hwnd, state ? SW_SHOW : SW_HIDE );
	}

	LONG getStyle()   { return GetWindowLong( hwnd, GWL_STYLE ); }
	LONG getExStyle() { return GetWindowLong( hwnd, GWL_EXSTYLE ); }

	void getWindowStyle( out LONG style, out LONG exstyle )
	{
		style   = getStyle();
		exstyle = getExStyle();
	}

	void adjustRect( inout RECT rect, bit hasMenu )
	{
		LONG style, exstyle;
		getWindowStyle( style, exstyle );
		AdjustWindowRectEx( &rect, style, hasMenu, exstyle );
	}

	void setWindowText( char[] text )
	{
		SetWindowText( hwnd, text );
	}

	void getWindowText( out char[] text )
	{
		int len;
		len = GetWindowTextLength( hwnd );
		text.length = len;
		GetWindowTextA( hwnd, text, len );
	}

	void moveWindow( RECT rect, bit repaint )
	{
	/*
		MoveWindow( hwnd,
						rect.left,
						rect.top,
						rect.right - rect.left,
						rect.bottom - rect.top,
						repaint );
	*/
		MoveWindow( hwnd,
						rect.left,
						rect.top,
						rect.right - rect.left,
						rect.bottom - rect.top,
						0 );
		InvalidateRect( hwnd, null, 0 );
	}

	void moveWindow( int x, int y, int w, int h, bit repaint )
	{
		MoveWindow( hwnd, x, y, w, h, repaint );
	}

	void getClientRect( out RECT rect )
	{
		GetClientRect( hwnd, &rect );
	}

	void getWindowRect( out RECT rect )
	{
		GetWindowRect( hwnd, &rect );
	}

	void screenToClient( inout POINT p )
	{
		ScreenToClient( hwnd, &p );
	}

	void screenToClient( inout int x, inout int y )
	{
		POINT p;
		p.x = x; p.y = y;
		ScreenToClient( hwnd, &p );
		x = p.x; y= p.y;
	}

	void clientToScreen( inout POINT p )
	{
		ClientToScreen( hwnd, &p );
	}

	void clientToScreen( inout int x, inout int y )
	{
		POINT p;
		p.x = x; p.y = y;
		ClientToScreen( hwnd, &p );
		x = p.x; y= p.y;
	}

	// get the windows rect in client coords.
	void getWindowBounds( out RECT rect )
	{
		GetWindowRect( hwnd, &rect );
		rect.right  -= rect.left;
		rect.bottom -= rect.top;
		rect.left  = 0;
		rect.top   = 0;
	}

	void getWindowPosition( out RECT rect )
	{
		POINT tl,br;
		HWND parentHwnd = GetParent( hwnd );
		GetWindowRect( hwnd, &rect );
		tl.x = rect.left;
		tl.y = rect.top;
		br.x = rect.right;
		br.y = rect.bottom;
		ScreenToClient( parentHwnd, &tl );
		ScreenToClient( parentHwnd, &br );
		rect.left   = tl.x;
		rect.top    = tl.y;
		rect.right  = br.x;
		rect.bottom = br.y;
	}

	void setSize( inout SIZE size )
	{
		RECT rect;
		POINT tl;
		GetWindowRect( hwnd, &rect );
		tl.x = rect.left;
		tl.y = rect.top;
		ScreenToClient( GetParent( hwnd ), &tl );
		MoveWindow( hwnd,	tl.x, tl.y, size.cx, size.cy, true );
	}
	
	void setParent( Win32Window newparent ) {
		SetParent( hwnd, newparent.hwnd );
	}

	Win32Window getParent() {
		return DFC.windowFromHandle( GetParent( hwnd ) );
	}

	void invalidate() {
		InvalidateRect( hwnd, null, false );
	}
}

class NativeWindow : Win32Window
{
public:
	this( char[] classname )
	{
		super( classname );
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int x, int y, int w, int h, HWND parent, HMENU menu )
	{
		CreateWindow(	style, windowName,styleEx,
							x, y, w, h,
							parent, menu);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, RECT * rect, HWND parent, HMENU menu )
	{
		CreateWindow(	style, windowName,	styleEx,
							rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top,
							parent,	menu);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int w, int h, HWND parent, HMENU menu )
	{
		CreateWindow(	style, windowName,styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, w, h,
							parent,	menu);
	}
	void create( DWORD style, DWORD styleEx, char[] windowName, SIZE * size, HWND parent, HMENU menu )
	{
		CreateWindow(	style, windowName, styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, size.cx, size.cy,
							parent,	menu);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int x, int y, int w, int h, HWND parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							x, y, w, h,
							parent,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, RECT * rect, HWND parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top,
							parent,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int w, int h, HWND parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, w, h,
							parent,	cast(HMENU)null);
	}
	void create( DWORD style, DWORD styleEx, char[] windowName, SIZE * size, HWND parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, size.cx, size.cy,
							parent,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int x, int y, int w, int h, NativeWindow parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							x, y, w, h,
							parent.hwnd,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, RECT * rect, NativeWindow parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top,
							parent.hwnd,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int w, int h, NativeWindow parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, w, h,
							parent.hwnd,	cast(HMENU)null);
	}
	void create( DWORD style, DWORD styleEx, char[] windowName, SIZE * size, NativeWindow parent )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, size.cx, size.cy,
							parent.hwnd,	cast(HMENU)null);
	}
	void create( DWORD style, DWORD styleEx, char[] windowName, int x, int y, int w, int h, NativeWindow parent, HMENU menu )
	{
		CreateWindow(	style, windowName,	styleEx,
							x, y, w, h,
							parent.hwnd,	menu);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, RECT * rect, NativeWindow parent, HMENU menu )
	{
		CreateWindow(	style, windowName,	styleEx,
							rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top,
							parent.hwnd,	menu);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int w, int h, NativeWindow parent, HMENU menu )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, w, h,
							parent.hwnd,	menu);
	}
	void create( DWORD style, DWORD styleEx, char[] windowName, SIZE * size, NativeWindow parent, HMENU menu )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, size.cx, size.cy,
							parent.hwnd,	menu);
	}
	//
	//
	//
	void create( DWORD style, DWORD styleEx, char[] windowName, int x, int y, int w, int h )
	{
		CreateWindow(	style, windowName,	styleEx,
							x, y, w, h,
							cast(HWND)null,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, RECT * rect )
	{
		CreateWindow(	style, windowName,	styleEx,
							rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top,
							cast(HWND)null,	cast(HMENU)null);
	}

	void create( DWORD style, DWORD styleEx, char[] windowName, int w, int h )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, w, h,
							cast(HWND)null,	cast(HMENU)null);
	}
	void create( DWORD style, DWORD styleEx, char[] windowName, SIZE * size )
	{
		CreateWindow(	style, windowName,	styleEx,
							CW_USEDEFAULT, CW_USEDEFAULT, size.cx, size.cy,
							cast(HWND)null,	cast(HMENU)null);
	}
}


//
// this is broken up to make the code a little more readable
//
class DWnd : NativeWindow
{
public:
	this( char[] classname ) 
	{ 
		super( classname ); 
	}
	this() { this(DFC.WINDCLASS); } 

	void addMouseDownHandler( UINT msg, mouseMsgFuncPtr func )
	{
		addHandler( msg, new MouseDownMessageSlot( hwnd, func ) );
	}

	void addMouseUpHandler( UINT msg, mouseMsgFuncPtr func )
	{
		addHandler( msg, new MouseUpMessageSlot( func ) );
	}

	void addMouseHandler( UINT msg, mouseMsgFuncPtr func )
	{
		addHandler( msg, new MouseMessageSlot( func ) );
	}

	void addKeyHandler( UINT msg, keyMsgFuncPtr func )
	{
		addHandler( msg, new KeyMessageSlot( func ) );
	}

	void addNativeHandler( UINT msg, nativeMsgFuncPtr func )
	{
		addHandler( msg, new NativeMessageSlot( func ) );
	}

	void addPaintHandler( UINT msg, paintMsgFuncPtr func )
	{
		addHandler( msg, new PaintMessageSlot( hwnd, func ) );
	}

	void addFocusHandler( UINT msg, focusMsgFuncPtr func )
	{
		addHandler( msg, new FocusMessageSlot( func ) );
	}
}


char[] getMsgName_prv( UINT msg )
{
	switch( msg )
	{
		case WM_NULL: return "NULL";
		case WM_CREATE: return "WM_CREATE";
		case WM_DESTROY: return "WM_DESTROY";
		case WM_MOVE: return "WM_MOVE";
		case WM_SIZE: return "WM_SIZE";

		case WM_ACTIVATE: return "WM_ACTIVATE";

		case WM_SETFOCUS: return "WM_SETFOCUS";
		case WM_KILLFOCUS: return "WM_KILLFOCUS";
		case WM_ENABLE: return "WM_ENABLE";
		case WM_SETREDRAW: return "WM_SETREDRAW";
		case WM_SETTEXT: return "WM_SETTEXT";
		case WM_GETTEXT: return "WM_GETTEXT";
		case WM_GETTEXTLENGTH: return "WM_GETTEXTLENGTH";
		case WM_PAINT: return "WM_PAINT";
		case WM_CLOSE: return "WM_CLOSE";
		case WM_QUERYENDSESSION: return "WM_QUERYENDSESSION";
		case WM_QUIT: return "WM_QUIT";
		case WM_QUERYOPEN: return "WM_QUERYOPEN";
		case WM_ERASEBKGND: return "WM_ERASEBKGND";
		case WM_SYSCOLORCHANGE: return "WM_SYSCOLORCHANGE";
		case WM_ENDSESSION: return "WM_ENDSESSION";
		case WM_SHOWWINDOW: return "WM_SHOWWINDOW";
		case WM_WININICHANGE: return "WM_WININICHANGE";
//		case WM_SETTINGCHANGE                =WM_WININICHANGE,

		case WM_DEVMODECHANGE: return "WM_DEVMODECHANGE";
		case WM_ACTIVATEAPP: return "WM_ACTIVATEAPP";
		case WM_FONTCHANGE: return "WM_FONTCHANGE";
		case WM_TIMECHANGE: return "WM_TIMECHANGE";
		case WM_CANCELMODE: return "WM_CANCELMODE";
		case WM_SETCURSOR: return "WM_SETCURSOR";
		case WM_MOUSEACTIVATE: return "WM_MOUSEACTIVATE";
		case WM_CHILDACTIVATE : return "WM_CHILDACTIVATE";
		case WM_QUEUESYNC: return "WM_QUEUESYNC";

		case WM_GETMINMAXINFO: return "WM_GETMINMAXINFO";

		case WM_PAINTICON: return "WM_PAINTICON";
		case WM_ICONERASEBKGND: return "WM_ICONERASEBKGND";
		case WM_NEXTDLGCTL: return "WM_NEXTDLGCTL";
		case WM_SPOOLERSTATUS: return "WM_SPOOLERSTATUS";
		case WM_DRAWITEM: return "WM_DRAWITEM";
		case WM_MEASUREITEM: return "WM_MEASUREITEM";
		case WM_DELETEITEM: return "WM_DELETEITEM";
		case WM_VKEYTOITEM: return "WM_VKEYTOITEM";
		case WM_CHARTOITEM: return "WM_CHARTOITEM";
		case WM_SETFONT: return "WM_SETFONT";
		case WM_GETFONT: return "WM_GETFONT";
		case WM_SETHOTKEY: return "WM_SETHOTKEY";
		case WM_GETHOTKEY: return "WM_GETHOTKEY";
		case WM_QUERYDRAGICON: return "WM_QUERYDRAGICON";
		case WM_COMPAREITEM: return "WM_COMPAREITEM";

		case WM_COMPACTING: return "WM_COMPACTING";
		case WM_COMMNOTIFY: return "WM_COMMNOTIFY";
		case WM_WINDOWPOSCHANGING: return "WM_WINDOWPOSCHANGING";
		case WM_WINDOWPOSCHANGED: return "WM_WINDOWPOSCHANGED";
		case WM_POWER: return "WM_POWER";

		case WM_COPYDATA: return "WM_COPYDATA";
		case WM_CANCELJOURNAL: return "WM_CANCELJOURNAL";
		case WM_NOTIFY: return "WM_NOTIFY";
		case WM_INPUTLANGCHANGEREQUEST: return "WM_INPUTLANGCHANGEREQUEST";
		case WM_INPUTLANGCHANGE: return "WM_INPUTLANGCHANGE";
		case WM_TCARD: return "WM_TCARD";
		case WM_HELP: return "WM_HELP";
		case WM_USERCHANGED: return "WM_USERCHANGED";
		case WM_NOTIFYFORMAT: return "WM_NOTIFYFORMAT";

		case WM_CONTEXTMENU: return "WM_CONTEXTMENU";
		case WM_STYLECHANGING: return "WM_STYLECHANGING";
		case WM_STYLECHANGED: return "WM_STYLECHANGED";
		case WM_DISPLAYCHANGE : return "WM_DISPLAYCHANGE";
		case WM_GETICON: return "WM_GETICON";
		case WM_SETICON: return "WM_SETICON";

		case WM_NCCREATE : return "WM_NCCREATE";
		case WM_NCDESTROY: return "WM_NCDESTROY";
		case WM_NCCALCSIZE : return "WM_NCCALCSIZE";
		case WM_NCHITTEST: return "WM_NCHITTEST";
		case WM_NCPAINT: return "WM_NCPAINT";
		case WM_NCACTIVATE: return "WM_NCACTIVATE";
		case WM_GETDLGCODE: return "WM_GETDLGCODE";

		case WM_NCMOUSEMOVE: return "WM_NCMOUSEMOVE";
		case WM_NCLBUTTONDOWN: return "WM_NCLBUTTONDOWN";
		case WM_NCLBUTTONUP: return "WM_NCLBUTTONUP";
		case WM_NCLBUTTONDBLCLK: return "WM_NCLBUTTONDBLCLK";
		case WM_NCRBUTTONDOWN: return "WM_NCRBUTTONDOWN";
		case WM_NCRBUTTONUP : return "WM_NCRBUTTONUP";
		case WM_NCRBUTTONDBLCLK: return "WM_NCRBUTTONDBLCLK";
		case WM_NCMBUTTONDOWN : return "WM_NCMBUTTONDOWN";
		case WM_NCMBUTTONUP: return "WM_NCMBUTTONUP";
		case WM_NCMBUTTONDBLCLK : return "WM_NCMBUTTONDBLCLK";

		case WM_KEYDOWN: return "WM_KEYDOWN";
		case WM_KEYUP: return "WM_KEYUP";
		case WM_CHAR: return "WM_CHAR";
		case WM_DEADCHAR: return "WM_DEADCHAR";
		case WM_SYSKEYDOWN: return "WM_SYSKEYDOWN";
		case WM_SYSKEYUP: return "WM_SYSKEYUP";
		case WM_SYSCHAR: return "WM_SYSCHAR";
		case WM_SYSDEADCHAR: return "WM_SYSDEADCHAR";
		case WM_KEYLAST: return "WM_KEYLAST";

		case WM_IME_STARTCOMPOSITION: return "WM_IME_STARTCOMPOSITION";
		case WM_IME_ENDCOMPOSITION: return "WM_IME_ENDCOMPOSITION";
		case WM_IME_COMPOSITION: return "WM_IME_COMPOSITION";

		case WM_INITDIALOG: return "WM_INITDIALOG";
		case WM_COMMAND: return "WM_COMMAND";
		case WM_SYSCOMMAND: return "WM_SYSCOMMAND";
		case WM_TIMER: return "WM_TIMER";
		case WM_HSCROLL: return "WM_HSCROLL";
		case WM_VSCROLL: return "WM_VSCROLL";
		case WM_INITMENU: return "WM_INITMENU";
		case WM_INITMENUPOPUP: return "WM_INITMENUPOPUP";
		case WM_MENUSELECT: return "WM_MENUSELECT";
		case WM_MENUCHAR: return "WM_MENUCHAR";
		case WM_ENTERIDLE: return "WM_ENTERIDLE";

		case WM_CTLCOLORMSGBOX: return "WM_CTLCOLORMSGBOX";
		case WM_CTLCOLOREDIT: return "WM_CTLCOLOREDIT";
		case WM_CTLCOLORLISTBOX: return "WM_CTLCOLORLISTBOX";
		case WM_CTLCOLORBTN: return "WM_CTLCOLORBTN";
		case WM_CTLCOLORDLG: return "WM_CTLCOLORDLG";
		case WM_CTLCOLORSCROLLBAR: return "WM_CTLCOLORSCROLLBAR";
		case WM_CTLCOLORSTATIC: return "WM_CTLCOLORSTATIC";

//		case WM_MOUSEFIRST: return "WM_MOUSEFIRST";
		case WM_MOUSEMOVE: return "WM_MOUSEMOVE";
		case WM_LBUTTONDOWN: return "WM_LBUTTONDOWN";
		case WM_LBUTTONUP: return "WM_LBUTTONUP";
		case WM_LBUTTONDBLCLK: return "WM_LBUTTONDBLCLK";
		case WM_RBUTTONDOWN: return "WM_RBUTTONDOWN";
		case WM_RBUTTONUP: return "WM_RBUTTONUP";
		case WM_RBUTTONDBLCLK: return "WM_RBUTTONDBLCLK";
		case WM_MBUTTONDOWN: return "WM_MBUTTONDOWN";
		case WM_MBUTTONUP: return "WM_MBUTTONUP";
		case WM_MBUTTONDBLCLK: return "WM_MBUTTONDBLCLK";
		case WM_MOUSEWHEEL: return "WM_MOUSEWHEEL";

		case WM_PARENTNOTIFY: return "WM_PARENTNOTIFY";
		case WM_ENTERMENULOOP: return "WM_ENTERMENULOOP";
		case WM_EXITMENULOOP: return "WM_EXITMENULOOP";

		case WM_NEXTMENU: return "WM_NEXTMENU";
		case WM_SIZING: return "WM_SIZING";
		case WM_CAPTURECHANGED: return "WM_CAPTURECHANGED";
		case WM_MOVING: return "WM_MOVING";
		case WM_POWERBROADCAST: return "WM_POWERBROADCAST";
		case WM_DEVICECHANGE: return "WM_DEVICECHANGE";

		case WM_IME_SETCONTEXT: return "WM_IME_SETCONTEXT";
		case WM_IME_NOTIFY: return "WM_IME_NOTIFY";
		case WM_IME_CONTROL: return "WM_IME_CONTROL";
		case WM_IME_COMPOSITIONFULL: return "WM_IME_COMPOSITIONFULL";
		case WM_IME_SELECT: return "WM_IME_SELECT";
		case WM_IME_CHAR: return "WM_IME_CHAR";
		case WM_IME_KEYDOWN: return "WM_IME_KEYDOWN";
		case WM_IME_KEYUP: return "WM_IME_KEYUP";

		case WM_MDICREATE: return "WM_MDICREATE";
		case WM_MDIDESTROY: return "WM_MDIDESTROY";
		case WM_MDIACTIVATE: return "WM_MDIACTIVATE";
		case WM_MDIRESTORE: return "WM_MDIRESTORE";
		case WM_MDINEXT: return "WM_MDINEXT";
		case WM_MDIMAXIMIZE: return "WM_MDIMAXIMIZE";
		case WM_MDITILE: return "WM_MDITILE";
		case WM_MDICASCADE: return "WM_MDICASCADE";
		case WM_MDIICONARRANGE: return "WM_MDIICONARRANGE";
		case WM_MDIGETACTIVE: return "WM_MDIGETACTIVE";

		case WM_MDISETMENU: return "WM_MDISETMENU";
		case WM_ENTERSIZEMOVE: return "WM_ENTERSIZEMOVE";
		case WM_EXITSIZEMOVE: return "WM_EXITSIZEMOVE";
		case WM_DROPFILES: return "WM_DROPFILES";
		case WM_MDIREFRESHMENU: return "WM_MDIREFRESHMENU";

		case WM_MOUSEHOVER: return "WM_MOUSEHOVER";
		case WM_MOUSELEAVE: return "WM_MOUSELEAVE";
		case WM_CUT: return "WM_CUT";
		case WM_COPY: return "WM_COPY";
		case WM_PASTE: return "WM_PASTE";
		case WM_CLEAR: return "WM_CLEAR";
		case WM_UNDO: return "WM_UNDO";
		case WM_RENDERFORMAT: return "WM_RENDERFORMAT";
		case WM_RENDERALLFORMATS: return "WM_RENDERALLFORMATS";
		case WM_DESTROYCLIPBOARD: return "WM_DESTROYCLIPBOARD";
		case WM_DRAWCLIPBOARD: return "WM_DRAWCLIPBOARD";
		case WM_PAINTCLIPBOARD: return "WM_PAINTCLIPBOARD";
		case WM_VSCROLLCLIPBOARD: return "WM_VSCROLLCLIPBOARD";
		case WM_SIZECLIPBOARD: return "WM_SIZECLIPBOARD";
		case WM_ASKCBFORMATNAME: return "WM_ASKCBFORMATNAME";
		case WM_CHANGECBCHAIN: return "WM_CHANGECBCHAIN";
		case WM_HSCROLLCLIPBOARD: return "WM_HSCROLLCLIPBOARD";
		case WM_QUERYNEWPALETTE: return "WM_QUERYNEWPALETTE";
		case WM_PALETTEISCHANGING: return "WM_PALETTEISCHANGING";
		case WM_PALETTECHANGED: return "WM_PALETTECHANGED";
		case WM_HOTKEY: return "WM_HOTKEY";
		case WM_PRINT: return "WM_PRINT";
		case WM_PRINTCLIENT: return "WM_PRINTCLIENT";

		case WM_HANDHELDFIRST: return "WM_HANDHELDFIRST";
		case WM_HANDHELDLAST: return "WM_HANDHELDLAST";

		case WM_AFXFIRST: return "WM_AFXFIRST";
		case WM_AFXLAST: return "WM_AFXLAST";

		case WM_PENWINFIRST: return "WM_PENWINFIRST";
		case WM_PENWINLAST: return "WM_PENWINLAST";

		default:
			if ( msg < 0x0400 ) { return "::reserved::"; }
			if ( msg >= 0x8000 ) { return "::application::"; }
			return "UserMsg";
	}
}

