module dfc.tree;

import std.c.stdio;

import win32.commctrl;
//import shellapi;
import dfc.core;

// D foundation Classes
//
// Win32 Tree control ...
//

/*
 * styles 
 *

TVS_HASBUTTONS         
TVS_HASLINES           
TVS_LINESATROOT        
TVS_EDITLABELS         
TVS_DISABLEDRAGDROP    
TVS_SHOWSELALWAYS      

 */

class DTreeCtrl : DSubClassedWnd
{
	this() 
	{ 
		super( WC_TREEVIEW ); 
	}

	HTREEITEM insertItem( LPTV_INSERTSTRUCT toInsert )
	{
		return cast(HTREEITEM)sendMsg( TVM_INSERTITEM, 0, cast(LPARAM)toInsert );
	}

	HTREEITEM insertItem( inout TV_INSERTSTRUCT toInsert )
	{
		return cast(HTREEITEM)sendMsg( TVM_INSERTITEM, 0, cast(LPARAM)&toInsert );
	}

	HTREEITEM insertItemA( LPTV_INSERTSTRUCTA toInsert )
	{
		printf( "inserting item via *\n" );
		return cast(HTREEITEM)sendMsg( TVM_INSERTITEMA, 0, cast(LPARAM)toInsert );
	}

	HTREEITEM insertItemA( inout TV_INSERTSTRUCTA toInsert )
	{
		printf( "inserting item via inout\n" );
		return cast(HTREEITEM)sendMsg( TVM_INSERTITEMA, 0, cast(LPARAM)&toInsert );
	}

	HTREEITEM insertItemW( LPTV_INSERTSTRUCTW toInsert )
	{
		return cast(HTREEITEM)sendMsg( TVM_INSERTITEMW, 0, cast(LPARAM)toInsert );
	}

	HTREEITEM insertItemW( inout TV_INSERTSTRUCTW toInsert )
	{
		return cast(HTREEITEM)sendMsg( TVM_INSERTITEMW, 0, cast(LPARAM)&toInsert );
	}

	HTREEITEM insertItem( inout TV_ITEM item, HTREEITEM parent, HTREEITEM insertAfter )
	{
		TV_INSERTSTRUCT toInsert;
		toInsert.hParent = parent;
		toInsert.hInsertAfter = insertAfter;
		toInsert.item = item;
		return insertItem( &toInsert );
	}

	HTREEITEM insertStringA( char[] str, HTREEITEM parent, HTREEITEM insertAfter )
	{
		printf( "importing string %s\n", cast(char*)str );

		TV_INSERTSTRUCTA toInsert;
		toInsert.hParent      = parent;
		toInsert.hInsertAfter = insertAfter;
		toInsert.item.mask           = TVIF_TEXT;
		toInsert.item.hItem          = null;
		toInsert.item.state          = 0;
		toInsert.item.stateMask      = 0;
		toInsert.item.pszText        = cast(char*)str;
		toInsert.item.cchTextMax     = 0;
		toInsert.item.iImage         = 0;
		toInsert.item.iSelectedImage = 0;
		toInsert.item.cChildren      = 0;
		toInsert.item.lParam         = 0;
		HTREEITEM ti = insertItemA( &toInsert );
		int i = cast(int)ti;
		printf( "insert item returned 0x%X\n", i );
		
		return ti;
	}

	HTREEITEM insertStringW( wchar[] str, HTREEITEM parent, HTREEITEM insertAfter )
	{
		TV_INSERTSTRUCTW toInsert;
		toInsert.hParent      = parent;
		toInsert.hInsertAfter = insertAfter;
		toInsert.item.mask           = TVIF_TEXT;
		toInsert.item.hItem          = null;
		toInsert.item.state          = 0;
		toInsert.item.stateMask      = 0;
		toInsert.item.pszText        = cast(wchar*)str;
		toInsert.item.cchTextMax     = 0;
		toInsert.item.iImage         = 0;
		toInsert.item.iSelectedImage = 0;
		toInsert.item.cChildren      = 0;
		toInsert.item.lParam         = 0;
		return insertItemW( &toInsert );
	}
// crashes the compiler (0.48 ALPHA)
//	version(UNICODE) {
//		alias insertStringW insertString;
//	} else {
//		alias insertStringA insertString;
//	}

	HTREEITEM insertString( char[] str, HTREEITEM parent, HTREEITEM insertAfter ) { return insertStringA( str, parent, insertAfter ); }

	bit deleteItem( HTREEITEM toDelete )
	{
		return (sendMsg( TVM_DELETEITEM, 0, cast(LPARAM)toDelete ) !=0);
	}

	bit deleteAllItems()
	{
		return (sendMsg( TVM_DELETEITEM, 0, cast(LPARAM)TVI_ROOT )!=0);
	}

	bit expand( HTREEITEM item, int flags )
	{
		return (sendMsg( TVM_EXPAND, cast(WPARAM)flags,cast(LPARAM)item )!=0);
	}

	bit getItemRect( inout RECT rect, HTREEITEM item, bit flags )
	{
		// this is realy how this gets the item handle!
//		*((HTREEITEM *)&rect) = item;
		return (sendMsg( TVM_GETITEMRECT, cast(WPARAM)flags,cast(LPARAM)&rect )!=0);
	}

	int getCount() 
	{
		return sendMsg( TVM_GETCOUNT, 0, 0 );
	}

	int getIndent() 
	{
		return sendMsg( TVM_GETINDENT, 0, 0 );
	}

	void setIndent( int indent ) 
	{
		sendMsg( TVM_SETINDENT, cast(WPARAM)indent, 0 );
	}
		
	HIMAGELIST getImageList( uint imageListToGet )
	{
		return cast(HIMAGELIST)sendMsg( TVM_GETIMAGELIST, cast(WPARAM)imageListToGet, 0 );
	}

	HIMAGELIST setImageList( HIMAGELIST imgList, uint imageListToGet )
	{
		return cast(HIMAGELIST)sendMsg( TVM_SETIMAGELIST, cast(WPARAM)imageListToGet, cast(LPARAM)imgList );
	}

	HTREEITEM getNextItem( HTREEITEM item, int code )
	{
		return cast(HTREEITEM)sendMsg( TVM_GETNEXTITEM, cast(WPARAM)code, cast(LPARAM)item );
	}

	HTREEITEM getChild( HTREEITEM item ) { 
		return getNextItem( item, TVGN_CHILD ); 
	}
	HTREEITEM getNextSibling( HTREEITEM item ) { 
		return getNextItem( item, TVGN_NEXT );
	}
	HTREEITEM getPrevSibling( HTREEITEM item ) {
		return getNextItem( item, TVGN_PREVIOUS );
	}
	HTREEITEM getParent( HTREEITEM item ) { 
		return getNextItem( item, TVGN_PARENT );
	}
	HTREEITEM getFirstVisible() { 
		return getNextItem( null, TVGN_FIRSTVISIBLE );
	}
	HTREEITEM getNextVisible( HTREEITEM item ) { 
		return getNextItem( item, TVGN_NEXTVISIBLE );
	}
	HTREEITEM getPrevVisible( HTREEITEM item ) { 
		return getNextItem( item, TVGN_PREVIOUSVISIBLE );
	}
	HTREEITEM getSelection() { 
		return getNextItem( null,  TVGN_CARET );
	}
	HTREEITEM getDropHilight() { 
		return getNextItem( null, TVGN_DROPHILITE );
	}
	HTREEITEM getRoot() { 
		return getNextItem( null, TVGN_ROOT ); 
	}

	HTREEITEM select( HTREEITEM item, int code )
	{
		return cast(HTREEITEM)sendMsg( TVM_SELECTITEM, cast(WPARAM)code, cast(LPARAM)item );
	}

	HTREEITEM selectItem( HTREEITEM item ) { 
		return select( item, TVGN_CARET ); 
	}
	HTREEITEM selectDropTarget( HTREEITEM item ) { 
		return select( item, TVGN_DROPHILITE );
	}
	HTREEITEM selectSetFirstVisible( HTREEITEM item ) { 
		return select( item, TVGN_FIRSTVISIBLE ); 
	}

	bit getItem( inout TV_ITEM item )
	{
		return (sendMsg( TVM_GETITEM, 0, cast(LPARAM)&item )!=0);
	}

	bit getItemA( inout TV_ITEM item )
	{
		return (sendMsg( TVM_GETITEMA, 0, cast(LPARAM)&item )!=0);
	}

	bit getItemW( inout TV_ITEM item )
	{
		return (sendMsg( TVM_GETITEMW, 0, cast(LPARAM)&item )!=0);
	}

	bit setItem( inout TV_ITEM item )
	{
		return (sendMsg( TVM_SETITEM, 0, cast(LPARAM)&item )!=0);
	}

	bit setItemA( inout TV_ITEM item )
	{
		return (sendMsg( TVM_SETITEMA, 0, cast(LPARAM)&item )!=0);
	}

	bit setItemW( inout TV_ITEM item )
	{
		return (sendMsg( TVM_SETITEMW, 0, cast(LPARAM)&item )!=0);
	}

	HWND editLabel( HTREEITEM item )
	{
		return cast(HWND)sendMsg( TVM_EDITLABEL, 0, cast(LPARAM)item );
	}

	HWND editLabelA( HTREEITEM item )
	{
		return cast(HWND)sendMsg( TVM_EDITLABELA, 0, cast(LPARAM)item );
	}

	HWND editLabelW( HTREEITEM item )
	{
		return cast(HWND)sendMsg( TVM_EDITLABELW, 0, cast(LPARAM)item );
	}

	HWND getEditControl()
	{
		return cast(HWND)sendMsg( TVM_GETEDITCONTROL, 0, 0 );
	}

	int getVisibleCount()
	{
		return sendMsg( TVM_GETVISIBLECOUNT, 0, 0 );
	}

	HTREEITEM hitTest( inout TV_HITTESTINFO htInfo )
	{
		return cast(HTREEITEM)sendMsg( TVM_HITTEST, 0, cast(LPARAM)&htInfo );
	}

	HIMAGELIST createDragImage( HTREEITEM item )
	{
		return cast(HIMAGELIST)sendMsg( TVM_CREATEDRAGIMAGE, 0, cast(LPARAM)item );
	}

	bit sortChildren( HTREEITEM item )
	{
		// recurse should be 0 (maybe Win2K does something not got the docs.
		return (sendMsg( TVM_SORTCHILDREN, 0, cast(LPARAM)item )!=0);
	}

	bit ensureVisible( HTREEITEM item )
	{
		return (sendMsg( TVM_ENSUREVISIBLE, 0, cast(LPARAM)item )!=0);
	}

	bit sortChildrenCB( inout TV_SORTCB cbStruct )
	{
		// recurse should be 0 (maybe Win2K does something not got the docs.
		return (sendMsg( TVM_SORTCHILDRENCB, 0, cast(LPARAM)&cbStruct )!=0);
	}

	bit endEditLabelNow( bit cancel )
	{
		return (sendMsg( TVM_ENDEDITLABELNOW, cast(WPARAM)cancel, 0 )!=0);
	}

	bit getISearchStringA( char[] str )
	{
		char * strPtr;
		strPtr = cast(char*)str;
		return (sendMsg( TVM_GETISEARCHSTRINGA, 0, cast(LPARAM)strPtr )!=0);
	}
		
	bit getISearchStringW( wchar[] str )
	{
		wchar * strPtr;
		strPtr = cast(wchar*)str;
		return (sendMsg( TVM_GETISEARCHSTRINGW, 0, cast(LPARAM)strPtr )!=0);
	}

	version(UNICODE) {
		alias getISearchStringW getISearchString;
	} else {
		alias getISearchStringA getISearchString;
	}
}

/** debug/hack/helper class */
class TreeHelper 
{
	static HIMAGELIST getSystemImageList()
	{

		return cast(HIMAGELIST)null;
	}
}
